# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test_numpyarray_merge():
    emptyarray = ak._v2.contents.EmptyArray()

    np1 = np.arange(2 * 7 * 5).reshape(2, 7, 5)
    np2 = np.arange(3 * 7 * 5).reshape(3, 7, 5)
    ak1 = ak._v2.contents.NumpyArray(np1)
    ak2 = ak._v2.contents.NumpyArray(np2)
    assert to_list(ak1.merge(ak2)) == to_list(np.concatenate([np1, np2]))
    assert to_list(ak1[1:, :-1, ::-1].merge(ak2[1:, :-1, ::-1])) == to_list(
        np.concatenate([np1[1:, :-1, ::-1], np2[1:, :-1, ::-1]])
    )
    assert ak1.typetracer.merge(ak2).form == ak1.merge(ak2).form
    assert (
        ak1[1:, :-1, ::-1].typetracer.merge(ak2[1:, :-1, ::-1]).form
        == ak1[1:, :-1, ::-1].merge(ak2[1:, :-1, ::-1]).form
    )

    for x in [
        np.bool_,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.float32,
        np.float64,
    ]:
        for y in [
            np.bool_,
            np.int8,
            np.int16,
            np.int32,
            np.int64,
            np.uint8,
            np.uint16,
            np.uint32,
            np.uint64,
            np.float32,
            np.float64,
        ]:
            z = np.concatenate(
                [np.array([1, 2, 3], dtype=x), np.array([4, 5], dtype=y)]
            ).dtype.type
            one = ak._v2.contents.NumpyArray(np.array([1, 2, 3], dtype=x))
            two = ak._v2.contents.NumpyArray(np.array([4, 5], dtype=y))
            three = one.merge(two)
            assert np.asarray(three).dtype == np.dtype(z), "{} {} {} {}".format(
                x, y, z, np.asarray(three).dtype.type
            )
            assert to_list(three) == to_list(
                np.concatenate([np.asarray(one), np.asarray(two)])
            )
            assert to_list(one.merge(emptyarray)) == to_list(one)
            assert to_list(emptyarray.merge(one)) == to_list(one)

            assert one.typetracer.merge(two).form == one.merge(two).form
            assert one.typetracer.merge(emptyarray).form == one.merge(emptyarray).form
            assert emptyarray.typetracer.merge(one).form == emptyarray.merge(one).form


def test_regulararray_merge():
    emptyarray = ak._v2.contents.EmptyArray()

    np1 = np.arange(2 * 7 * 5).reshape(2, 7, 5)
    np2 = np.arange(3 * 7 * 5).reshape(3, 7, 5)
    ak1 = ak._v2.operations.convert.from_iter(np1, highlevel=False)
    ak2 = ak._v2.operations.convert.from_iter(np2, highlevel=False)

    assert to_list(ak1.merge(ak2)) == to_list(np.concatenate([np1, np2]))
    assert to_list(ak1.merge(emptyarray)) == to_list(ak1)
    assert to_list(emptyarray.merge(ak1)) == to_list(ak1)

    assert ak1.typetracer.merge(ak2).form == ak1.merge(ak2).form
    assert ak1.typetracer.merge(emptyarray).form == ak1.merge(emptyarray).form
    assert emptyarray.typetracer.merge(ak1).form == emptyarray.merge(ak1).form


def test_listarray_merge():
    emptyarray = ak._v2.contents.EmptyArray()

    content1 = ak._v2.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    content2 = ak._v2.contents.NumpyArray(np.array([1, 2, 3, 4, 5, 6, 7]))

    for (dtype1, Index1, ListArray1), (dtype2, Index2, ListArray2) in [
        (
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
        ),
        (
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
        ),
        (
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
        ),
        (
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
        ),
        (
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
        ),
        (
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
        ),
        (
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
        ),
        (
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
        ),
        (
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
        ),
    ]:
        starts1 = Index1(np.array([0, 3, 3], dtype=dtype1))
        stops1 = Index1(np.array([3, 3, 5], dtype=dtype1))
        starts2 = Index2(np.array([2, 99, 0], dtype=dtype2))
        stops2 = Index2(np.array([6, 99, 3], dtype=dtype2))
        array1 = ListArray1(starts1, stops1, content1)
        array2 = ListArray2(starts2, stops2, content2)
        assert to_list(array1) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
        assert to_list(array2) == [[3, 4, 5, 6], [], [1, 2, 3]]

        assert to_list(array1.merge(array2)) == [
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
            [3, 4, 5, 6],
            [],
            [1, 2, 3],
        ]
        assert to_list(array2.merge(array1)) == [
            [3, 4, 5, 6],
            [],
            [1, 2, 3],
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
        ]
        assert to_list(array1.merge(emptyarray)) == to_list(array1)
        assert to_list(emptyarray.merge(array1)) == to_list(array1)

        assert array1.typetracer.merge(array2).form == array1.merge(array2).form
        assert array2.typetracer.merge(array1).form == array2.merge(array1).form
        assert array1.typetracer.merge(emptyarray).form == array1.merge(emptyarray).form
        assert emptyarray.typetracer.merge(array1).form == emptyarray.merge(array1).form

    regulararray = ak._v2.contents.RegularArray(content2, 2, zeros_length=0)
    assert to_list(regulararray) == [[1, 2], [3, 4], [5, 6]]
    assert to_list(regulararray.merge(emptyarray)) == to_list(regulararray)
    assert to_list(emptyarray.merge(regulararray)) == to_list(regulararray)

    for (dtype1, Index1, ListArray1) in [
        (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
        (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
        (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
    ]:
        starts1 = Index1(np.array([0, 3, 3], dtype=dtype1))
        stops1 = Index1(np.array([3, 3, 5], dtype=dtype1))
        array1 = ListArray1(starts1, stops1, content1)

        assert to_list(array1.merge(regulararray)) == [
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
            [1, 2],
            [3, 4],
            [5, 6],
        ]
        assert to_list(regulararray.merge(array1)) == [
            [1, 2],
            [3, 4],
            [5, 6],
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
        ]


def test_listoffsetarray_merge():
    emptyarray = ak._v2.contents.EmptyArray()

    content1 = ak._v2.contents.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))
    content2 = ak._v2.contents.NumpyArray(np.array([1, 2, 3, 4, 5, 6, 7]))

    for (dtype1, Index1, ListOffsetArray1), (dtype2, Index2, ListOffsetArray2) in [
        (
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListOffsetArray),
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListOffsetArray),
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListOffsetArray),
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListOffsetArray),
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListOffsetArray),
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListOffsetArray),
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray),
            (np.int32, ak._v2.index.Index32, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray),
            (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListOffsetArray),
        ),
        (
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray),
            (np.int64, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray),
        ),
    ]:
        offsets1 = Index1(np.array([0, 3, 3, 5], dtype=dtype1))
        offsets2 = Index2(np.array([1, 3, 3, 3, 5], dtype=dtype2))
        array1 = ListOffsetArray1(offsets1, content1)
        array2 = ListOffsetArray2(offsets2, content2)
        assert to_list(array1) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
        assert to_list(array2) == [[2, 3], [], [], [4, 5]]

        assert to_list(array1.merge(array2)) == [
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
            [2, 3],
            [],
            [],
            [4, 5],
        ]
        assert to_list(array2.merge(array1)) == [
            [2, 3],
            [],
            [],
            [4, 5],
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
        ]
        assert to_list(array1.merge(emptyarray)) == to_list(array1)
        assert to_list(emptyarray.merge(array1)) == to_list(array1)

        assert array1.typetracer.merge(array2).form == array1.merge(array2).form
        assert array2.typetracer.merge(array1).form == array2.merge(array1).form
        assert array1.typetracer.merge(emptyarray).form == array1.merge(emptyarray).form
        assert emptyarray.typetracer.merge(array1).form == emptyarray.merge(array1).form

    regulararray = ak._v2.contents.RegularArray(content2, 2, zeros_length=0)
    assert to_list(regulararray) == [[1, 2], [3, 4], [5, 6]]

    for (dtype1, Index1, ListArray1) in [
        (np.int32, ak._v2.index.Index32, ak._v2.contents.ListArray),
        (np.uint32, ak._v2.index.IndexU32, ak._v2.contents.ListArray),
        (np.int64, ak._v2.index.Index64, ak._v2.contents.ListArray),
    ]:
        starts1 = Index1(np.array([0, 3, 3], dtype=dtype1))
        stops1 = Index1(np.array([3, 3, 5], dtype=dtype1))
        array1 = ListArray1(starts1, stops1, content1)

        assert to_list(array1.merge(regulararray)) == [
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
            [1, 2],
            [3, 4],
            [5, 6],
        ]
        assert to_list(regulararray.merge(array1)) == [
            [1, 2],
            [3, 4],
            [5, 6],
            [1.1, 2.2, 3.3],
            [],
            [4.4, 5.5],
        ]

        assert (
            array1.typetracer.merge(regulararray).form
            == array1.merge(regulararray).form
        )
        assert (
            regulararray.typetracer.merge(array1).form
            == regulararray.merge(array1).form
        )


def test_recordarray_merge():
    emptyarray = ak._v2.contents.EmptyArray()

    arrayr1 = ak._v2.operations.convert.from_iter(
        [{"x": 0, "y": []}, {"x": 1, "y": [1, 1]}, {"x": 2, "y": [2, 2]}],
        highlevel=False,
    )
    arrayr2 = ak._v2.operations.convert.from_iter(
        [
            {"x": 2.2, "y": [2.2, 2.2]},
            {"x": 1.1, "y": [1.1, 1.1]},
            {"x": 0.0, "y": [0.0, 0.0]},
        ],
        highlevel=False,
    )
    arrayr3 = ak._v2.operations.convert.from_iter(
        [{"x": 0, "y": 0.0}, {"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}],
        highlevel=False,
    )
    arrayr4 = ak._v2.operations.convert.from_iter(
        [{"y": [], "x": 0}, {"y": [1, 1], "x": 1}, {"y": [2, 2], "x": 2}],
        highlevel=False,
    )
    arrayr5 = ak._v2.operations.convert.from_iter(
        [
            {"x": 0, "y": [], "z": 0},
            {"x": 1, "y": [1, 1], "z": 1},
            {"x": 2, "y": [2, 2], "z": 2},
        ],
        highlevel=False,
    )
    arrayr6 = ak._v2.operations.convert.from_iter(
        [
            {"z": 0, "x": 0, "y": []},
            {"z": 1, "x": 1, "y": [1, 1]},
            {"z": 2, "x": 2, "y": [2, 2]},
        ],
        highlevel=False,
    )
    arrayr7 = ak._v2.operations.convert.from_iter(
        [{"x": 0}, {"x": 1}, {"x": 2}], highlevel=False
    )

    arrayt1 = ak._v2.operations.convert.from_iter(
        [(0, []), (1, [1.1]), (2, [2, 2])], highlevel=False
    )
    arrayt2 = ak._v2.operations.convert.from_iter(
        [(2.2, [2.2, 2.2]), (1.1, [1.1, 1.1]), (0.0, [0.0, 0.0])], highlevel=False
    )
    arrayt3 = ak._v2.operations.convert.from_iter(
        [(0, 0.0), (1, 1.1), (2, 2.2)], highlevel=False
    )
    arrayt4 = ak._v2.operations.convert.from_iter(
        [([], 0), ([1.1], 1), ([2.2, 2.2], 2)], highlevel=False
    )
    arrayt5 = ak._v2.operations.convert.from_iter(
        [(0, [], 0), (1, [1], 1), (2, [2, 2], 2)], highlevel=False
    )
    arrayt6 = ak._v2.operations.convert.from_iter(
        [(0, 0, []), (1, 1, [1]), (2, 2, [2, 2])], highlevel=False
    )
    arrayt7 = ak._v2.operations.convert.from_iter([(0,), (1,), (2,)], highlevel=False)

    assert arrayr1.mergeable(arrayr2)
    assert arrayr2.mergeable(arrayr1)
    assert not arrayr1.mergeable(arrayr3)
    assert arrayr1.mergeable(arrayr4)
    assert arrayr4.mergeable(arrayr1)
    assert not arrayr1.mergeable(arrayr5)
    assert not arrayr1.mergeable(arrayr6)
    assert arrayr5.mergeable(arrayr6)
    assert arrayr6.mergeable(arrayr5)
    assert not arrayr1.mergeable(arrayr7)

    assert arrayt1.mergeable(arrayt2)
    assert arrayt2.mergeable(arrayt1)
    assert not arrayt1.mergeable(arrayt3)
    assert not arrayt1.mergeable(arrayt4)
    assert not arrayt1.mergeable(arrayt5)
    assert not arrayt1.mergeable(arrayt6)
    assert not arrayt5.mergeable(arrayt6)
    assert not arrayt1.mergeable(arrayt7)

    assert to_list(arrayr1.merge(arrayr2)) == [
        {"x": 0.0, "y": []},
        {"x": 1.0, "y": [1.0, 1.0]},
        {"x": 2.0, "y": [2.0, 2.0]},
        {"x": 2.2, "y": [2.2, 2.2]},
        {"x": 1.1, "y": [1.1, 1.1]},
        {"x": 0.0, "y": [0.0, 0.0]},
    ]
    assert to_list(arrayr2.merge(arrayr1)) == [
        {"x": 2.2, "y": [2.2, 2.2]},
        {"x": 1.1, "y": [1.1, 1.1]},
        {"x": 0.0, "y": [0.0, 0.0]},
        {"x": 0.0, "y": []},
        {"x": 1.0, "y": [1.0, 1.0]},
        {"x": 2.0, "y": [2.0, 2.0]},
    ]

    assert to_list(arrayr1.merge(arrayr4)) == [
        {"x": 0, "y": []},
        {"x": 1, "y": [1, 1]},
        {"x": 2, "y": [2, 2]},
        {"x": 0, "y": []},
        {"x": 1, "y": [1, 1]},
        {"x": 2, "y": [2, 2]},
    ]
    assert to_list(arrayr4.merge(arrayr1)) == [
        {"x": 0, "y": []},
        {"x": 1, "y": [1, 1]},
        {"x": 2, "y": [2, 2]},
        {"x": 0, "y": []},
        {"x": 1, "y": [1, 1]},
        {"x": 2, "y": [2, 2]},
    ]

    assert to_list(arrayr5.merge(arrayr6)) == [
        {"x": 0, "y": [], "z": 0},
        {"x": 1, "y": [1, 1], "z": 1},
        {"x": 2, "y": [2, 2], "z": 2},
        {"x": 0, "y": [], "z": 0},
        {"x": 1, "y": [1, 1], "z": 1},
        {"x": 2, "y": [2, 2], "z": 2},
    ]
    assert to_list(arrayr6.merge(arrayr5)) == [
        {"x": 0, "y": [], "z": 0},
        {"x": 1, "y": [1, 1], "z": 1},
        {"x": 2, "y": [2, 2], "z": 2},
        {"x": 0, "y": [], "z": 0},
        {"x": 1, "y": [1, 1], "z": 1},
        {"x": 2, "y": [2, 2], "z": 2},
    ]

    assert to_list(arrayt1.merge(arrayt2)) == [
        (0.0, []),
        (1.0, [1.1]),
        (2.0, [2.0, 2.0]),
        (2.2, [2.2, 2.2]),
        (1.1, [1.1, 1.1]),
        (0.0, [0.0, 0.0]),
    ]
    assert to_list(arrayt2.merge(arrayt1)) == [
        (2.2, [2.2, 2.2]),
        (1.1, [1.1, 1.1]),
        (0.0, [0.0, 0.0]),
        (0.0, []),
        (1.0, [1.1]),
        (2.0, [2.0, 2.0]),
    ]

    assert arrayr1.typetracer.merge(arrayr2).form == arrayr1.merge(arrayr2).form
    assert arrayr2.typetracer.merge(arrayr1).form == arrayr2.merge(arrayr1).form
    assert arrayr1.typetracer.merge(arrayr4).form == arrayr1.merge(arrayr4).form
    assert arrayr4.typetracer.merge(arrayr1).form == arrayr4.merge(arrayr1).form
    assert arrayr5.typetracer.merge(arrayr6).form == arrayr5.merge(arrayr6).form
    assert arrayr6.typetracer.merge(arrayr5).form == arrayr6.merge(arrayr5).form
    assert arrayt1.typetracer.merge(arrayt2).form == arrayt1.merge(arrayt2).form
    assert arrayt2.typetracer.merge(arrayt1).form == arrayt2.merge(arrayt1).form

    assert to_list(arrayr1.merge(emptyarray)) == to_list(arrayr1)
    assert to_list(arrayr2.merge(emptyarray)) == to_list(arrayr2)
    assert to_list(arrayr3.merge(emptyarray)) == to_list(arrayr3)
    assert to_list(arrayr4.merge(emptyarray)) == to_list(arrayr4)
    assert to_list(arrayr5.merge(emptyarray)) == to_list(arrayr5)
    assert to_list(arrayr6.merge(emptyarray)) == to_list(arrayr6)
    assert to_list(arrayr7.merge(emptyarray)) == to_list(arrayr7)

    assert to_list(emptyarray.merge(arrayr1)) == to_list(arrayr1)
    assert to_list(emptyarray.merge(arrayr2)) == to_list(arrayr2)
    assert to_list(emptyarray.merge(arrayr3)) == to_list(arrayr3)
    assert to_list(emptyarray.merge(arrayr4)) == to_list(arrayr4)
    assert to_list(emptyarray.merge(arrayr5)) == to_list(arrayr5)
    assert to_list(emptyarray.merge(arrayr6)) == to_list(arrayr6)
    assert to_list(emptyarray.merge(arrayr7)) == to_list(arrayr7)

    assert to_list(arrayt1.merge(emptyarray)) == to_list(arrayt1)
    assert to_list(arrayt2.merge(emptyarray)) == to_list(arrayt2)
    assert to_list(arrayt3.merge(emptyarray)) == to_list(arrayt3)
    assert to_list(arrayt4.merge(emptyarray)) == to_list(arrayt4)
    assert to_list(arrayt5.merge(emptyarray)) == to_list(arrayt5)
    assert to_list(arrayt6.merge(emptyarray)) == to_list(arrayt6)
    assert to_list(arrayt7.merge(emptyarray)) == to_list(arrayt7)

    assert to_list(emptyarray.merge(arrayt1)) == to_list(arrayt1)
    assert to_list(emptyarray.merge(arrayt2)) == to_list(arrayt2)
    assert to_list(emptyarray.merge(arrayt3)) == to_list(arrayt3)
    assert to_list(emptyarray.merge(arrayt4)) == to_list(arrayt4)
    assert to_list(emptyarray.merge(arrayt5)) == to_list(arrayt5)
    assert to_list(emptyarray.merge(arrayt6)) == to_list(arrayt6)
    assert to_list(emptyarray.merge(arrayt7)) == to_list(arrayt7)

    assert arrayr1.typetracer.merge(emptyarray).form == arrayr1.merge(emptyarray).form
    assert arrayr2.typetracer.merge(emptyarray).form == arrayr2.merge(emptyarray).form
    assert arrayr3.typetracer.merge(emptyarray).form == arrayr3.merge(emptyarray).form
    assert arrayr4.typetracer.merge(emptyarray).form == arrayr4.merge(emptyarray).form
    assert arrayr5.typetracer.merge(emptyarray).form == arrayr5.merge(emptyarray).form
    assert arrayr6.typetracer.merge(emptyarray).form == arrayr6.merge(emptyarray).form
    assert arrayr7.typetracer.merge(emptyarray).form == arrayr7.merge(emptyarray).form

    assert emptyarray.typetracer.merge(arrayr1).form == emptyarray.merge(arrayr1).form
    assert emptyarray.typetracer.merge(arrayr2).form == emptyarray.merge(arrayr2).form
    assert emptyarray.typetracer.merge(arrayr3).form == emptyarray.merge(arrayr3).form
    assert emptyarray.typetracer.merge(arrayr4).form == emptyarray.merge(arrayr4).form
    assert emptyarray.typetracer.merge(arrayr5).form == emptyarray.merge(arrayr5).form
    assert emptyarray.typetracer.merge(arrayr6).form == emptyarray.merge(arrayr6).form
    assert emptyarray.typetracer.merge(arrayr7).form == emptyarray.merge(arrayr7).form

    assert arrayt1.typetracer.merge(emptyarray).form == arrayt1.merge(emptyarray).form
    assert arrayt2.typetracer.merge(emptyarray).form == arrayt2.merge(emptyarray).form
    assert arrayt3.typetracer.merge(emptyarray).form == arrayt3.merge(emptyarray).form
    assert arrayt4.typetracer.merge(emptyarray).form == arrayt4.merge(emptyarray).form
    assert arrayt5.typetracer.merge(emptyarray).form == arrayt5.merge(emptyarray).form
    assert arrayt6.typetracer.merge(emptyarray).form == arrayt6.merge(emptyarray).form
    assert arrayt7.typetracer.merge(emptyarray).form == arrayt7.merge(emptyarray).form

    assert emptyarray.typetracer.merge(arrayt1).form == emptyarray.merge(arrayt1).form
    assert emptyarray.typetracer.merge(arrayt2).form == emptyarray.merge(arrayt2).form
    assert emptyarray.typetracer.merge(arrayt3).form == emptyarray.merge(arrayt3).form
    assert emptyarray.typetracer.merge(arrayt4).form == emptyarray.merge(arrayt4).form
    assert emptyarray.typetracer.merge(arrayt5).form == emptyarray.merge(arrayt5).form
    assert emptyarray.typetracer.merge(arrayt6).form == emptyarray.merge(arrayt6).form
    assert emptyarray.typetracer.merge(arrayt7).form == emptyarray.merge(arrayt7).form


def test_indexedarray_merge():
    content1 = ak._v2.operations.convert.from_iter(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False
    )
    content2 = ak._v2.operations.convert.from_iter(
        [[1, 2], [], [3, 4]], highlevel=False
    )
    index1 = ak._v2.index.Index64(np.array([2, 0, -1, 0, 1, 2], dtype=np.int64))
    indexedarray1 = ak._v2.contents.IndexedOptionArray(index1, content1)

    assert to_list(indexedarray1) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]

    assert to_list(indexedarray1.merge(content2)) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [1.0, 2.0],
        [],
        [3.0, 4.0],
    ]
    assert to_list(content2.merge(indexedarray1)) == [
        [1.0, 2.0],
        [],
        [3.0, 4.0],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert to_list(indexedarray1.merge(indexedarray1)) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        None,
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]

    assert (
        indexedarray1.typetracer.merge(content2).form
        == indexedarray1.merge(content2).form
    )
    assert (
        content2.typetracer.merge(indexedarray1).form
        == content2.merge(indexedarray1).form
    )
    assert (
        indexedarray1.typetracer.merge(indexedarray1).form
        == indexedarray1.merge(indexedarray1).form
    )


def test_unionarray_merge():
    emptyarray = ak._v2.contents.EmptyArray()

    one = ak._v2.operations.convert.from_iter(
        [0.0, 1.1, 2.2, [], [1], [2, 2]], highlevel=False
    )
    two = ak._v2.operations.convert.from_iter(
        [{"x": 1, "y": 1.1}, 999, 123, {"x": 2, "y": 2.2}], highlevel=False
    )
    three = ak._v2.operations.convert.from_iter(
        ["one", "two", "three"], highlevel=False
    )

    assert to_list(one.merge(two)) == [
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
    ]
    assert to_list(two.merge(one)) == [
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
    ]

    assert to_list(one.merge(emptyarray)) == [0.0, 1.1, 2.2, [], [1], [2, 2]]
    assert to_list(emptyarray.merge(one)) == [0.0, 1.1, 2.2, [], [1], [2, 2]]

    assert to_list(one.merge(three)) == [
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
        "one",
        "two",
        "three",
    ]
    assert to_list(two.merge(three)) == [
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
        "one",
        "two",
        "three",
    ]
    assert to_list(three.merge(one)) == [
        "one",
        "two",
        "three",
        0.0,
        1.1,
        2.2,
        [],
        [1],
        [2, 2],
    ]
    assert to_list(three.merge(two)) == [
        "one",
        "two",
        "three",
        {"x": 1, "y": 1.1},
        999,
        123,
        {"x": 2, "y": 2.2},
    ]

    assert one.typetracer.merge(two).form == one.merge(two).form
    assert two.typetracer.merge(one).form == two.merge(one).form
    assert one.typetracer.merge(emptyarray).form == one.merge(emptyarray).form
    assert emptyarray.typetracer.merge(one).form == emptyarray.merge(one).form
    assert one.typetracer.merge(three).form == one.merge(three).form
    assert two.typetracer.merge(three).form == two.merge(three).form
    assert three.typetracer.merge(one).form == three.merge(one).form
    assert three.typetracer.merge(two).form == three.merge(two).form


def test_merge_parameters():
    one = ak._v2.operations.convert.from_iter(
        [[121, 117, 99, 107, 121], [115, 116, 117, 102, 102]], highlevel=False
    )
    two = ak._v2.operations.convert.from_iter(["good", "stuff"], highlevel=False)

    assert to_list(ak._v2.operations.structure.concatenate([one, two])) == [
        [121, 117, 99, 107, 121],
        [115, 116, 117, 102, 102],
        "good",
        "stuff",
    ]
    assert to_list(ak._v2.operations.structure.concatenate([two, one])) == [
        "good",
        "stuff",
        [121, 117, 99, 107, 121],
        [115, 116, 117, 102, 102],
    ]

    assert (
        ak._v2.operations.structure.concatenate(
            [one, two], highlevel=False
        ).typetracer.form
        == ak._v2.operations.structure.concatenate([one, two], highlevel=False).form
    )
    assert (
        ak._v2.operations.structure.concatenate(
            [two, one], highlevel=False
        ).typetracer.form
        == ak._v2.operations.structure.concatenate([two, one], highlevel=False).form
    )


def test_mask_as_bool():
    array = ak._v2.operations.convert.from_iter(
        ["one", "two", None, "three", None, None, "four"], highlevel=False
    )
    index2 = ak._v2.index.Index64(np.array([2, 2, 1, 5, 0], dtype=np.int64))
    array2 = ak._v2.contents.IndexedArray(index2, array)
    assert np.asarray(array.mask_as_bool(valid_when=False).view(np.int8)).tolist() == [
        0,
        0,
        1,
        0,
        1,
        1,
        0,
    ]
    assert np.asarray(array2.mask_as_bool(valid_when=False).view(np.int8)).tolist() == [
        0,
        0,
        0,
        0,
        0,
    ]


def test_indexedarray_simplify():
    array = ak._v2.operations.convert.from_iter(
        ["one", "two", None, "three", None, None, "four", "five"], highlevel=False
    )
    index2 = ak._v2.index.Index64(np.array([2, 2, 1, 6, 5], dtype=np.int64))

    array2 = ak._v2.contents.IndexedArray(index2, array)
    assert np.asarray(array.index).tolist() == [0, 1, -1, 2, -1, -1, 3, 4]
    assert (
        to_list(array2.simplify_optiontype())
        == to_list(array2)
        == [None, None, "two", "four", None]
    )

    assert (
        array2.typetracer.simplify_optiontype().form
        == array2.simplify_optiontype().form
    )


def test_indexedarray_simplify_more():
    content = ak._v2.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )

    index1_32 = ak._v2.index.Index32(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.int32))
    index1_U32 = ak._v2.index.IndexU32(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.uint32))
    index1_64 = ak._v2.index.Index64(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.int64))
    index2_32 = ak._v2.index.Index32(np.array([0, 2, 4, 6], dtype=np.int32))
    index2_U32 = ak._v2.index.IndexU32(np.array([0, 2, 4, 6], dtype=np.uint32))
    index2_64 = ak._v2.index.Index64(np.array([0, 2, 4, 6], dtype=np.int64))

    array = ak._v2.contents.IndexedArray(
        index2_32, ak._v2.contents.IndexedArray(index1_32, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_32, ak._v2.contents.IndexedArray(index1_U32, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_32, ak._v2.contents.IndexedArray(index1_64, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_U32, ak._v2.contents.IndexedArray(index1_32, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_U32, ak._v2.contents.IndexedArray(index1_U32, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_U32, ak._v2.contents.IndexedArray(index1_64, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_64, ak._v2.contents.IndexedArray(index1_32, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_64, ak._v2.contents.IndexedArray(index1_U32, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_64, ak._v2.contents.IndexedArray(index1_64, content)
    )
    assert to_list(array) == [6.6, 4.4, 2.2, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, 4.4, 2.2, 0.0]
    assert isinstance(
        array.simplify_optiontype(), ak._v2.contents.indexedarray.IndexedArray
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    index1_32 = ak._v2.index.Index32(np.array([6, 5, -1, 3, -1, 1, 0], dtype=np.int32))
    index1_64 = ak._v2.index.Index64(np.array([6, 5, -1, 3, -1, 1, 0], dtype=np.int64))
    index2_32 = ak._v2.index.Index32(np.array([0, 2, 4, 6], dtype=np.int32))
    index2_U32 = ak._v2.index.IndexU32(np.array([0, 2, 4, 6], dtype=np.uint32))
    index2_64 = ak._v2.index.Index64(np.array([0, 2, 4, 6], dtype=np.int64))

    array = ak._v2.contents.IndexedArray(
        index2_32, ak._v2.contents.IndexedOptionArray(index1_32, content)
    )
    assert to_list(array) == [6.6, None, None, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, 0.0]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_32, ak._v2.contents.IndexedOptionArray(index1_64, content)
    )
    assert to_list(array) == [6.6, None, None, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, 0.0]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_U32, ak._v2.contents.IndexedOptionArray(index1_32, content)
    )
    assert to_list(array) == [6.6, None, None, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, 0.0]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_U32, ak._v2.contents.IndexedOptionArray(index1_64, content)
    )
    assert to_list(array) == [6.6, None, None, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, 0.0]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_64, ak._v2.contents.IndexedOptionArray(index1_32, content)
    )
    assert to_list(array) == [6.6, None, None, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, 0.0]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedArray(
        index2_64, ak._v2.contents.IndexedOptionArray(index1_64, content)
    )
    assert to_list(array) == [6.6, None, None, 0.0]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, 0.0]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    index1_32 = ak._v2.index.Index32(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.int32))
    index1_U32 = ak._v2.index.IndexU32(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.uint32))
    index1_64 = ak._v2.index.Index64(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.int64))
    index2_32 = ak._v2.index.Index32(np.array([0, -1, 4, -1], dtype=np.int32))
    index2_64 = ak._v2.index.Index64(np.array([0, -1, 4, -1], dtype=np.int64))

    array = ak._v2.contents.IndexedOptionArray(
        index2_32, ak._v2.contents.IndexedArray(index1_32, content)
    )
    assert to_list(array) == [6.6, None, 2.2, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, 2.2, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_32, ak._v2.contents.IndexedArray(index1_U32, content)
    )
    assert to_list(array) == [6.6, None, 2.2, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, 2.2, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_32, ak._v2.contents.IndexedArray(index1_64, content)
    )
    assert to_list(array) == [6.6, None, 2.2, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, 2.2, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_64, ak._v2.contents.IndexedArray(index1_32, content)
    )
    assert to_list(array) == [6.6, None, 2.2, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, 2.2, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_64, ak._v2.contents.IndexedArray(index1_U32, content)
    )
    assert to_list(array) == [6.6, None, 2.2, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, 2.2, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_64, ak._v2.contents.IndexedArray(index1_64, content)
    )
    assert to_list(array) == [6.6, None, 2.2, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, 2.2, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    index1_32 = ak._v2.index.Index32(np.array([6, 5, -1, 3, -1, 1, 0], dtype=np.int32))
    index1_64 = ak._v2.index.Index64(np.array([6, 5, -1, 3, -1, 1, 0], dtype=np.int64))
    index2_32 = ak._v2.index.Index32(np.array([0, -1, 4, -1], dtype=np.int32))
    index2_64 = ak._v2.index.Index64(np.array([0, -1, 4, -1], dtype=np.int64))

    array = ak._v2.contents.IndexedOptionArray(
        index2_32, ak._v2.contents.IndexedOptionArray(index1_32, content)
    )
    assert to_list(array) == [6.6, None, None, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_32, ak._v2.contents.IndexedOptionArray(index1_64, content)
    )
    assert to_list(array) == [6.6, None, None, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_64, ak._v2.contents.IndexedOptionArray(index1_32, content)
    )
    assert to_list(array) == [6.6, None, None, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )

    array = ak._v2.contents.IndexedOptionArray(
        index2_64, ak._v2.contents.IndexedOptionArray(index1_64, content)
    )
    assert to_list(array) == [6.6, None, None, None]
    assert to_list(array.simplify_optiontype()) == [6.6, None, None, None]
    assert isinstance(
        array.simplify_optiontype(),
        ak._v2.contents.indexedoptionarray.IndexedOptionArray,
    )
    assert isinstance(
        array.simplify_optiontype().content, ak._v2.contents.numpyarray.NumpyArray
    )
    assert (
        array.typetracer.simplify_optiontype().form == array.simplify_optiontype().form
    )


def test_unionarray_simplify_one():
    one = ak._v2.operations.convert.from_iter([5, 4, 3, 2, 1], highlevel=False)
    two = ak._v2.operations.convert.from_iter(
        [[], [1], [2, 2], [3, 3, 3]], highlevel=False
    )
    three = ak._v2.operations.convert.from_iter([1.1, 2.2, 3.3], highlevel=False)
    tags = ak._v2.index.Index8(
        np.array([0, 0, 1, 2, 1, 0, 2, 1, 1, 0, 2, 0], dtype=np.int8)
    )
    index = ak._v2.index.Index64(
        np.array([0, 1, 0, 0, 1, 2, 1, 2, 3, 3, 2, 4], dtype=np.int64)
    )
    array = ak._v2.contents.UnionArray(tags, index, [one, two, three])

    assert to_list(array) == [
        5,
        4,
        [],
        1.1,
        [1],
        3,
        2.2,
        [2, 2],
        [3, 3, 3],
        2,
        3.3,
        1,
    ]
    assert to_list(array.simplify_uniontype(True, False)) == [
        5.0,
        4.0,
        [],
        1.1,
        [1],
        3.0,
        2.2,
        [2, 2],
        [3, 3, 3],
        2.0,
        3.3,
        1.0,
    ]
    assert len(array.contents) == 3
    assert len(array.simplify_uniontype(True, False).contents) == 2

    assert (
        array.typetracer.simplify_uniontype(True, False).form
        == array.simplify_uniontype(True, False).form
    )


def test_unionarray_simplify():
    one = ak._v2.operations.convert.from_iter([5, 4, 3, 2, 1], highlevel=False)
    two = ak._v2.operations.convert.from_iter(
        [[], [1], [2, 2], [3, 3, 3]], highlevel=False
    )
    three = ak._v2.operations.convert.from_iter([1.1, 2.2, 3.3], highlevel=False)

    tags2 = ak._v2.index.Index8(np.array([0, 1, 0, 1, 0, 0, 1], dtype=np.int8))
    index2 = ak._v2.index.Index32(np.array([0, 0, 1, 1, 2, 3, 2], dtype=np.int32))
    inner = ak._v2.contents.UnionArray(tags2, index2, [two, three])
    tags1 = ak._v2.index.Index8(
        np.array([0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 1, 0], dtype=np.int8)
    )
    index1 = ak._v2.index.Index64(
        np.array([0, 1, 0, 1, 2, 2, 3, 4, 5, 3, 6, 4], dtype=np.int64)
    )
    outer = ak._v2.contents.UnionArray(tags1, index1, [one, inner])
    assert to_list(outer) == [
        5,
        4,
        [],
        1.1,
        [1],
        3,
        2.2,
        [2, 2],
        [3, 3, 3],
        2,
        3.3,
        1,
    ]

    assert to_list(outer.simplify_uniontype(True, False)) == [
        5.0,
        4.0,
        [],
        1.1,
        [1],
        3.0,
        2.2,
        [2, 2],
        [3, 3, 3],
        2.0,
        3.3,
        1.0,
    ]
    assert isinstance(outer.content(1), ak._v2.contents.unionarray.UnionArray)
    assert isinstance(
        outer.simplify_uniontype(True, False).content(0),
        ak._v2.contents.numpyarray.NumpyArray,
    )
    assert isinstance(
        outer.simplify_uniontype(True, False).content(1),
        ak._v2.contents.listoffsetarray.ListOffsetArray,
    )
    assert len(outer.simplify_uniontype(True, False).contents) == 2
    assert (
        outer.typetracer.simplify_uniontype(True, False).form
        == outer.simplify_uniontype(True, False).form
    )

    tags2 = ak._v2.index.Index8(np.array([0, 1, 0, 1, 0, 0, 1], dtype=np.int8))
    index2 = ak._v2.index.Index64(np.array([0, 0, 1, 1, 2, 3, 2], dtype=np.int64))
    inner = ak._v2.contents.UnionArray(tags2, index2, [two, three])
    tags1 = ak._v2.index.Index8(
        np.array([1, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1], dtype=np.int8)
    )
    index1 = ak._v2.index.Index32(
        np.array([0, 1, 0, 1, 2, 2, 3, 4, 5, 3, 6, 4], dtype=np.int32)
    )
    outer = ak._v2.contents.UnionArray(tags1, index1, [inner, one])
    assert to_list(outer) == [
        5,
        4,
        [],
        1.1,
        [1],
        3,
        2.2,
        [2, 2],
        [3, 3, 3],
        2,
        3.3,
        1,
    ]


def test_concatenate():
    one = ak._v2.highlevel.Array([1.1, 2.2, 3.3, 4.4, 5.5], check_valid=True).layout
    two = ak._v2.highlevel.Array([[], [1], [2, 2], [3, 3, 3]], check_valid=True).layout
    three = ak._v2.highlevel.Array(
        [True, False, False, True, True], check_valid=True
    ).layout

    assert to_list(ak._v2.operations.structure.concatenate([one, two, three])) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        [],
        [1],
        [2, 2],
        [3, 3, 3],
        1.0,
        0.0,
        0.0,
        1.0,
        1.0,
    ]
    assert isinstance(
        ak._v2.operations.structure.concatenate([one, two, three], highlevel=False),
        ak._v2.contents.unionarray.UnionArray,
    )
    assert (
        len(
            ak._v2.operations.structure.concatenate(
                [one, two, three], highlevel=False
            ).contents
        )
        == 2
    )


def test_where():
    condition = ak._v2.highlevel.Array(
        [True, False, True, False, True],
        check_valid=True,
    )
    one = ak._v2.highlevel.Array([1.1, 2.2, 3.3, 4.4, 5.5], check_valid=True)
    two = ak._v2.highlevel.Array([False, False, False, True, True], check_valid=True)
    three = ak._v2.highlevel.Array(
        [[], [1], [2, 2], [3, 3, 3], [4, 4, 4, 4]], check_valid=True
    )

    assert to_list(ak._v2.operations.structure.where(condition, one, two)) == [
        1.1,
        0.0,
        3.3,
        1.0,
        5.5,
    ]
    assert to_list(ak._v2.operations.structure.where(condition, one, three)) == [
        1.1,
        [1],
        3.3,
        [3, 3, 3],
        5.5,
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


class Canary(dict):
    def __init__(self):
        super().__init__()
        self.ops = []

    def __getitem__(self, key):
        self.ops.append(("get", key))
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        self.ops.append(("set", key))
        return super().__setitem__(key, value)


def test_lazy_buffers():

    array = ak.from_json(
        """
    [
        {
            "listcollection": [
                {"item1": 1, "item2": 2},
                {"item1": 2, "item2": 4},
                {"item1": 3, "item2": 6}
            ],
            "collection": {"item1": 3, "item2": 4},
            "singleton": 5,
            "listsingleton": [1, 2, 3],
            "unioncollection": {"item1": 3},
            "masked": null
        },
        {
            "listcollection": [
                {"item1": 1, "item2": 2},
                {"item1": 2, "item2": 4},
                {"item1": 3, "item2": 6}
            ],
            "collection": {"item1": 3, "item2": 4},
            "singleton": 5,
            "listsingleton": [1, 2, 3],
            "unioncollection": [{"item1": 2}],
            "masked": 4
        },
        {
            "listcollection": [
                {"item1": 1, "item2": 2},
                {"item1": 2, "item2": 4},
                {"item1": 3, "item2": 6}
            ],
            "collection": {"item1": 3, "item2": 4},
            "singleton": 5,
            "listsingleton": [1, 2, 3],
            "unioncollection": {"item1": 4},
            "masked": 4
        }
    ]"""
    )

    canary = Canary()
    key_format = "kitty-{form_key}-{attribute}"
    form, length, container = ak.to_buffers(
        array, container=canary, key_format=key_format
    )
    assert not any(op[0] == "get" for op in canary.ops)
    canary.ops = []

    cache = {}
    out = ak.from_buffers(
        form,
        length,
        container,
        key_format=key_format,
        lazy=True,
        lazy_cache=cache,
        lazy_cache_key="hello",
    )
    assert len(canary.ops) == 0
    assert len(cache) == 0

    assert len(out) == 3
    assert len(canary.ops) == 0
    assert len(cache) == 0

    assert ak.to_list(ak.num(out.listcollection)) == [3, 3, 3]
    assert set(canary.ops) == {("get", "kitty-node1-offsets")}
    assert "hello" in cache
    assert any(x.startswith("hello(kitty-node1-virtual") for x in cache)
    canary.ops = []
    cache.clear()

    assert ak.to_list(out.unioncollection) == [
        {"item1": 3},
        [{"item1": 2}],
        {"item1": 4},
    ]
    assert set(canary.ops) == {
        ("get", "kitty-node11-tags"),
        ("get", "kitty-node11-index"),
        ("get", "kitty-node14-offsets"),
        ("get", "kitty-node13-data"),
        ("get", "kitty-node16-data"),
    }
    assert "hello" in cache
    assert any(x.startswith("hello(kitty-node11-virtual") for x in cache)
    assert any(x.startswith("hello(kitty-node13-virtual") for x in cache)
    assert any(x.startswith("hello(kitty-node16-virtual") for x in cache)
    canary.ops = []
    cache.clear()

    assert ak.to_list(out.masked) == [None, 4, 4]
    assert set(canary.ops) == {
        ("get", "kitty-node17-index"),
        ("get", "kitty-node18-data"),
    }
    assert "hello" in cache
    assert any(x.startswith("hello(kitty-node17-virtual") for x in cache)
    canary.ops = []
    cache.clear()


def test_longer_than_expected():
    array = ak.Array(
        ak.layout.ListOffsetArray64(
            ak.layout.Index64([0, 2, 4]),
            ak.layout.RecordArray(
                {
                    "item1": ak.layout.NumpyArray(np.arange(4)),
                    "longitem": ak.layout.NumpyArray(np.arange(6)),
                }
            ),
        )
    )
    out = ak.from_buffers(*ak.to_buffers(array), lazy=True)
    assert ak.to_list(out) == [
        [{"item1": 0, "longitem": 0}, {"item1": 1, "longitem": 1}],
        [{"item1": 2, "longitem": 2}, {"item1": 3, "longitem": 3}],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

# v2: keep this file, but modify it to only get objects that exist!

# NumPy-like alternatives
import awkward.nplike

# shims for C++ (now everything is compiled into one 'awkward._ext' module)
import awkward.layout
import awkward.types
import awkward.forms
import awkward.partition

# internal
import awkward._v2
import awkward._cpu_kernels
import awkward._libawkward
import awkward._util

# third-party connectors
import awkward._connect._numpy
import awkward._connect._numba
import awkward._connect._numexpr
import awkward._connect._autograd

# high-level interface
behavior = {}
from awkward.highlevel import Array
from awkward.highlevel import Record
from awkward.highlevel import ArrayBuilder

# third-party jax connectors
import awkward._connect._jax

# behaviors
from awkward.behaviors.mixins import *
from awkward.behaviors.string import *
from awkward.behaviors.categorical import *

# operations
from awkward.operations.convert import *
from awkward.operations.describe import *
from awkward.operations.structure import *
from awkward.operations.reducers import *

# version
__version__ = awkward._ext.__version__

# call C++ startup function
awkward._ext.startup()

__all__ = [x for x in list(globals()) if not x.startswith("_") and x not in ("numpy",)]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    ak_array = ak.Array([1, 2, 3])
    assert ak.singletons(ak_array).tolist() == [[1], [2], [3]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import copy

import awkward as ak
from awkward._v2.contents.content import Content
from awkward._v2.forms.unmaskedform import UnmaskedForm
from awkward._v2.forms.form import _parameters_equal

np = ak.nplike.NumpyMetadata.instance()


class UnmaskedArray(Content):
    is_OptionType = True

    def __init__(self, content, identifier=None, parameters=None, nplike=None):
        if not isinstance(content, Content):
            raise TypeError(
                "{} 'content' must be a Content subtype, not {}".format(
                    type(self).__name__, repr(content)
                )
            )
        if nplike is None:
            nplike = content.nplike

        self._content = content
        self._init(identifier, parameters, nplike)

    @property
    def content(self):
        return self._content

    Form = UnmaskedForm

    def _form_with_key(self, getkey):
        form_key = getkey(self)
        return self.Form(
            self._content._form_with_key(getkey),
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(self, form, getkey, container, nplike):
        assert isinstance(form, self.Form)
        self._content._to_buffers(form.content, getkey, container, nplike)

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return UnmaskedArray(
            self._content.typetracer,
            self._typetracer_identifier(),
            self._parameters,
            tt,
        )

    @property
    def length(self):
        return self._content.length

    def _forget_length(self):
        return UnmaskedArray(
            self._content._forget_length(),
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<UnmaskedArray len="]
        out.append(repr(str(self.length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</UnmaskedArray>")
        out.append(post)
        return "".join(out)

    def merge_parameters(self, parameters):
        return UnmaskedArray(
            self._content,
            self._identifier,
            ak._v2._util.merge_parameters(self._parameters, parameters),
            self._nplike,
        )

    def toByteMaskedArray(self):
        return ak._v2.contents.bytemaskedarray.ByteMaskedArray(
            ak._v2.index.Index8(self.mask_as_bool(valid_when=True).view(np.int8)),
            self._content,
            True,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def toIndexedOptionArray64(self):
        arange = self._nplike.arange(self._content.length, dtype=np.int64)
        return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
            ak._v2.index.Index64(arange),
            self._content,
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def mask_as_bool(self, valid_when=True, nplike=None):
        if nplike is None:
            nplike = self._nplike

        if valid_when:
            return nplike.ones(self._content.length, dtype=np.bool_)
        else:
            return nplike.zeros(self._content.length, dtype=np.bool_)

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if not self._nplike.known_data:
            return ak._v2._typetracer.MaybeNone(self._content._getitem_at(where))

        return self._content._getitem_at(where)

    def _getitem_range(self, where):
        if not self._nplike.known_shape:
            return self

        start, stop, step = where.indices(self.length)
        assert step == 1
        return UnmaskedArray(
            self._content._getitem_range(slice(start, stop)),
            self._range_identifier(start, stop),
            self._parameters,
            self._nplike,
        )

    def _getitem_field(self, where, only_fields=()):
        return UnmaskedArray(
            self._content._getitem_field(where, only_fields),
            self._field_identifier(where),
            None,
            self._nplike,
        )

    def _getitem_fields(self, where, only_fields=()):
        return UnmaskedArray(
            self._content._getitem_fields(where, only_fields),
            self._fields_identifier(where),
            None,
            self._nplike,
        )

    def _carry(self, carry, allow_lazy, exception):
        return UnmaskedArray(
            self._content._carry(carry, allow_lazy, exception),
            self._carry_identifier(carry, exception),
            self._parameters,
            self._nplike,
        )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, (int, slice, ak._v2.index.Index64)):
            return UnmaskedArray(
                self._content._getitem_next(head, tail, advanced),
                self._identifier,
                self._parameters,
                self._nplike,
            ).simplify_optiontype()

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            raise NotImplementedError

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            raise NotImplementedError

        else:
            raise AssertionError(repr(head))

    def project(self, mask=None):
        if mask is not None:
            return ak._v2.contents.bytemaskedarray.ByteMaskedArray(
                mask,
                self._content,
                False,
                None,
                self._parameters,
                self._nplike,
            ).project()
        else:
            return self._content

    def simplify_optiontype(self):
        if isinstance(
            self._content,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return self._content
        else:
            return self

    def num(self, axis, depth=0):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            out = ak._v2.index.Index64.empty(1, self._nplike)
            out[0] = self.length
            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
                0
            ]
        else:
            return ak._v2.contents.unmaskedarray.UnmaskedArray(
                self._content.num(posaxis, depth), None, self._parameters, self._nplike
            )

    def _offsets_and_flattened(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            raise np.AxisError(self, "axis=0 not allowed for flatten")
        else:
            offsets, flattened = self._content._offsets_and_flattened(posaxis, depth)
            if offsets.length == 0:
                return (
                    offsets,
                    UnmaskedArray(flattened, None, self._parameters, self._nplike),
                )

            else:
                return (offsets, flattened)

    def mergeable(self, other, mergebool):
        if not _parameters_equal(self._parameters, other._parameters):
            return False

        if isinstance(
            other,
            (
                ak._v2.contents.emptyarray.EmptyArray,
                ak._v2.contents.unionarray.UnionArray,
            ),
        ):
            return True

        if isinstance(
            other,
            (
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return self._content.mergeable(other.content, mergebool)

        else:
            return self._content.mergeable(other, mergebool)

    def _reverse_merge(self, other):
        return self.toIndexedOptionArray64()._reverse_merge(other)

    def _mergemany(self, others):
        if len(others) == 0:
            return self
        return self.toIndexedOptionArray64().mergemany(others)

    def fillna(self, value):
        return self._content.fillna(value)

    def _localindex(self, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        else:
            return UnmaskedArray(
                self._content._localindex(posaxis, depth),
                self._identifier,
                self._parameters,
                self._nplike,
            )

    def numbers_to_type(self, name):
        return ak._v2.contents.unmaskedarray.UnmaskedArray(
            self._content.numbers_to_type(name),
            self._identifier,
            self._parameters,
            self._nplike,
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        if self._content.length == 0:
            return True
        return self._content._is_unique(negaxis, starts, parents, outlength)

    def _unique(self, negaxis, starts, parents, outlength):
        if self._content.length == 0:
            return self
        return self._content._unique(negaxis, starts, parents, outlength)

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self._content.length == 0:
            return ak._v2.contents.NumpyArray(
                self._nplike.empty(0, np.int64), None, None, self._nplike
            )

        out = self._content._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

        if isinstance(out, ak._v2.contents.RegularArray):
            tmp = ak._v2.contents.UnmaskedArray(
                out._content,
                None,
                None,
                self._nplike,
            ).simplify_optiontype()

            return ak._v2.contents.RegularArray(
                tmp,
                out._size,
                out._length,
                None,
                None,
                self._nplike,
            )

        else:
            return out

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        out = self._content._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

        if isinstance(out, ak._v2.contents.RegularArray):
            tmp = ak._v2.contents.UnmaskedArray(
                out._content,
                self._identifier,
                self._parameters,
                self._nplike,
            ).simplify_optiontype()

            return ak._v2.contents.RegularArray(
                tmp,
                out._size,
                out._length,
                self._identifier,
                self._parameters,
                self._nplike,
            )

        else:
            return out

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            return ak._v2.contents.unmaskedarray.UnmaskedArray(
                self._content._combinations(
                    n, replacement, recordlookup, parameters, posaxis, depth
                ),
                self._identifier,
                self._parameters,
                self._nplike,
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        next = self._content
        if isinstance(next, ak._v2_contents.RegularArray):
            next = next.toListOffsetArray64(True)

        return next._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        if isinstance(
            self._content,
            (
                ak._v2.contents.bitmaskedarray.BitMaskedArray,
                ak._v2.contents.bytemaskedarray.ByteMaskedArray,
                ak._v2.contents.indexedarray.IndexedArray,
                ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                ak._v2.contents.unmaskedarray.UnmaskedArray,
            ),
        ):
            return "{0} contains \"{1}\", the operation that made it might have forgotten to call 'simplify_optiontype()'"
        else:
            return self._content.validityerror(path + ".content")

    def _nbytes_part(self):
        result = self.content._nbytes_part()
        if self.identifier is not None:
            result = result + self.identifier._nbytes_part()
        return result

    def _rpad(self, target, axis, depth, clip):
        posaxis = self.axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self.rpad_axis0(target, clip)
        elif posaxis == depth + 1:
            return self._content._rpad(target, posaxis, depth, clip)
        else:
            return ak._v2.contents.unmaskedarray.UnmaskedArray(
                self._content._rpad(target, posaxis, depth, clip),
                None,
                self._parameters,
                self._nplike,
            )

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        return self._content._to_arrow(pyarrow, self, None, length, options)

    def _to_numpy(self, allow_missing):
        content = ak._v2.operations.convert.to_numpy(
            self.content, allow_missing=allow_missing
        )
        if allow_missing:
            return self._nplike.ma.MaskedArray(content)
        else:
            return content

    def _completely_flatten(self, nplike, options):
        return self.project()._completely_flatten(nplike, options)

    def _recursively_apply(
        self, action, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:

            def continuation():
                return UnmaskedArray(
                    self._content._recursively_apply(
                        action,
                        depth,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    ),
                    self._identifier,
                    self._parameters if options["keep_parameters"] else None,
                    self._nplike,
                )

        else:

            def continuation():
                self._content._recursively_apply(
                    action,
                    depth,
                    copy.copy(depth_context),
                    lateral_context,
                    options,
                )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def packed(self):
        return UnmaskedArray(
            self._content.packed(), self._identifier, self._parameters, self._nplike
        )

    def _to_list(self, behavior):
        out = self._to_list_custom(behavior)
        if out is not None:
            return out

        return self._content._to_list(behavior)

    def _to_nplike(self, nplike):
        content = self._content._to_nplike(nplike)
        return UnmaskedArray(
            content,
            identifier=self.identifier,
            parameters=self.parameters,
            nplike=nplike,
        )

    def _to_json(
        self,
        nan_string,
        infinity_string,
        minus_infinity_string,
        complex_real_string,
        complex_imag_string,
    ):
        out = self._to_json_custom()
        if out is not None:
            return out

        return self._content._to_json(
            nan_string,
            infinity_string,
            minus_infinity_string,
            complex_real_string,
            complex_imag_string,
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list

empty = ak._v2.highlevel.Array(
    ak._v2.contents.RegularArray(
        ak._v2.highlevel.Array([[1, 2, 3], [], [4, 5]]).layout, 0, zeros_length=0
    )
)


def test_ListOffsetArray_rpad_and_clip():
    array = ak._v2.highlevel.Array([[1, 2, 3], [], [4, 5]])
    assert ak._v2.operations.structure.pad_none(array, 0, clip=True).tolist() == [
        [],
        [],
        [],
    ]

    array = ak._v2.highlevel.Array([[1, 2, 3], [], [4, 5]])
    assert ak._v2.operations.structure.pad_none(array, 0).tolist() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]


def test_toListOffsetArray64():
    assert ak._v2.operations.structure.from_regular(empty).tolist() == []


def test_carry():
    assert empty[[]].tolist() == []


def test_num():
    assert ak._v2.operations.structure.num(empty, axis=0) == 0
    assert ak._v2.operations.structure.num(empty, axis=1).tolist() == []
    assert ak._v2.operations.structure.num(empty, axis=2).tolist() == []


def test_flatten():
    assert ak._v2.operations.structure.flatten(empty, axis=0).tolist() == []
    assert ak._v2.operations.structure.flatten(empty, axis=1).tolist() == []
    assert ak._v2.operations.structure.flatten(empty, axis=2).tolist() == []


def test_mergeable():
    assert ak._v2.operations.structure.concatenate([empty, empty]).tolist() == []


def test_fillna():
    assert ak._v2.operations.structure.fill_none(empty, 5, axis=0).tolist() == []


def test_pad_none():
    assert ak._v2.operations.structure.pad_none(empty, 0, axis=0).tolist() == []
    assert ak._v2.operations.structure.pad_none(empty, 0, axis=1).tolist() == []
    assert ak._v2.operations.structure.pad_none(empty, 0, axis=2).tolist() == []

    assert ak._v2.operations.structure.pad_none(empty, 1, axis=0).tolist() == [None]
    assert ak._v2.operations.structure.pad_none(empty, 1, axis=1).tolist() == []
    assert ak._v2.operations.structure.pad_none(empty, 1, axis=2).tolist() == []

    assert (
        ak._v2.operations.structure.pad_none(empty, 0, axis=0, clip=True).tolist() == []
    )
    assert (
        ak._v2.operations.structure.pad_none(empty, 0, axis=1, clip=True).tolist() == []
    )
    assert (
        ak._v2.operations.structure.pad_none(empty, 0, axis=2, clip=True).tolist() == []
    )

    assert ak._v2.operations.structure.pad_none(
        empty, 1, axis=0, clip=True
    ).tolist() == [None]
    assert (
        ak._v2.operations.structure.pad_none(empty, 1, axis=1, clip=True).tolist() == []
    )
    assert (
        ak._v2.operations.structure.pad_none(empty, 1, axis=2, clip=True).tolist() == []
    )


def test_reduce():
    assert ak._v2.operations.reducers.sum(empty, axis=0).tolist() == []


def test_localindex():
    assert ak._v2.operations.structure.local_index(empty, axis=0).tolist() == []
    assert ak._v2.operations.structure.local_index(empty, axis=1).tolist() == []
    assert ak._v2.operations.structure.local_index(empty, axis=2).tolist() == []


def test_combinations():
    assert ak._v2.operations.structure.combinations(empty, 2, axis=0).tolist() == []
    assert ak._v2.operations.structure.combinations(empty, 2, axis=1).tolist() == []
    assert ak._v2.operations.structure.combinations(empty, 2, axis=2).tolist() == []


def test_getitem():
    with pytest.raises(IndexError):
        empty[
            0,
        ]

    jagged = ak._v2.highlevel.Array([[]])[0:0]
    assert empty[jagged].tolist() == []



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak
from awkward._v2.forms.form import Form, _parameters_equal


class UnmaskedForm(Form):
    is_OptionType = True

    def __init__(
        self,
        content,
        has_identifier=False,
        parameters=None,
        form_key=None,
    ):
        if not isinstance(content, Form):
            raise TypeError(
                "{} all 'contents' must be Form subclasses, not {}".format(
                    type(self).__name__, repr(content)
                )
            )

        self._content = content
        self._init(has_identifier, parameters, form_key)

    @property
    def content(self):
        return self._content

    def __repr__(self):
        args = [repr(self._content)] + self._repr_args()
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _tolist_part(self, verbose, toplevel):
        return self._tolist_extra(
            {
                "class": "UnmaskedArray",
                "content": self._content._tolist_part(verbose, toplevel=False),
            },
            verbose,
        )

    def _type(self, typestrs):
        return ak._v2.types.optiontype.OptionType(
            self._content._type(typestrs),
            self._parameters,
            ak._v2._util.gettypestr(self._parameters, typestrs),
        ).simplify_option_union()

    def __eq__(self, other):
        if isinstance(other, UnmaskedForm):
            return (
                self._has_identifier == other._has_identifier
                and self._form_key == other._form_key
                and _parameters_equal(self._parameters, other._parameters)
                and self._content == other._content
            )
        else:
            return False

    def generated_compatibility(self, other):
        if other is None:
            return True

        elif isinstance(other, UnmaskedForm):
            return _parameters_equal(
                self._parameters, other._parameters
            ) and self._content.generated_compatibility(other._content)

        else:
            return False

    def _getitem_range(self):
        return UnmaskedForm(
            self._content._getitem_range(),
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def _getitem_field(self, where, only_fields=()):
        return UnmaskedForm(
            self._content._getitem_field(where, only_fields),
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return UnmaskedForm(
            self._content._getitem_fields(where, only_fields),
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _carry(self, allow_lazy):
        return UnmaskedForm(
            self._content._carry(allow_lazy),
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def simplify_optiontype(self):
        if isinstance(
            self._content,
            (
                ak._v2.forms.indexedform.IndexedForm,
                ak._v2.forms.indexedoptionform.IndexedOptionForm,
                ak._v2.forms.bytemaskedform.ByteMaskedForm,
                ak._v2.forms.bitmaskedform.BitMaskedForm,
                ak._v2.forms.unmaskedform.UnmaskedForm,
            ),
        ):
            return self._content
        else:
            return self

    def purelist_parameter(self, key):
        if self._parameters is None or key not in self._parameters:
            return self._content.purelist_parameter(key)
        else:
            return self._parameters[key]

    @property
    def purelist_isregular(self):
        return self._content.purelist_isregular

    @property
    def purelist_depth(self):
        return self._content.purelist_depth

    @property
    def minmax_depth(self):
        return self._content.minmax_depth

    @property
    def branch_depth(self):
        return self._content.branch_depth

    @property
    def fields(self):
        return self._content.fields

    @property
    def dimension_optiontype(self):
        return True



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def parameters(array):

    """
    Extracts parameters from the outermost array node of `array` (many types
    supported, including all Awkward Arrays and Records).

    Parameters are a dict from str to JSON-like objects, usually strings.
    Every #ak.layout.Content node has a different set of parameters. Some
    key names are special, such as `"__record__"` and `"__array__"` that name
    particular records and arrays as capable of supporting special behaviors.

    See #ak.Array and #ak.behavior for a more complete description of
    behaviors.
    """
    if isinstance(array, (ak._v2.highlevel.Array, ak._v2.highlevel.Record)):
        return array.layout.parameters

    elif isinstance(
        array,
        (ak._v2.contents.Content, ak._v2.record.Record),
    ):
        return array.parameters

    elif isinstance(array, ak._v2.highlevel.ArrayBuilder):
        return array.snapshot().layout.parameters

    elif isinstance(array, ak.layout.ArrayBuilder):
        return array.snapshot().parameters

    else:
        return {}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import base64
import struct
import json
import re

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# C++17 is required for optional, variant
headers = ("stdexcept", "string", "optional", "variant", "complex", "chrono")
cache = {}


def generate_ArrayView(compiler, use_cached=True):
    key = "ArrayView"
    if use_cached:
        out = cache.get(key)
    else:
        out = None

    if out is None:
        out = """
namespace awkward {
  class ArrayView {
  public:
    ArrayView(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : start_(start), stop_(stop), which_(which), ptrs_(ptrs) { }

    size_t size() const noexcept {{
      return stop_ - start_;
    }}

    bool empty() const noexcept {{
      return start_ == stop_;
    }}

  protected:
    ssize_t start_;
    ssize_t stop_;
    ssize_t which_;
    ssize_t* ptrs_;
  };
}
""".strip()
        cache[key] = out
        compiler(out)

    return out


def generate_RecordView(compiler, use_cached=True):
    key = "RecordView"
    if use_cached:
        out = cache.get(key)
    else:
        out = None

    if out is None:
        out = """
namespace awkward {
  class RecordView {
  public:
    RecordView(ssize_t at, ssize_t which, ssize_t* ptrs)
      : at_(at), which_(which), ptrs_(ptrs) { }

  protected:
    ssize_t at_;
    ssize_t which_;
    ssize_t* ptrs_;
  };
}
""".strip()
        cache[key] = out
        compiler(out)

    return out


def togenerator(form):
    if isinstance(form, ak._v2.forms.EmptyForm):
        return togenerator(form.toNumpyForm(np.dtype(np.float64)))

    elif isinstance(form, ak._v2.forms.NumpyForm):
        if len(form.inner_shape) == 0:
            return NumpyArrayGenerator.from_form(form)
        else:
            return togenerator(form.toRegularForm())

    elif isinstance(form, ak._v2.forms.RegularForm):
        return RegularArrayGenerator.from_form(form)

    elif isinstance(form, (ak._v2.forms.ListForm, ak._v2.forms.ListOffsetForm)):
        return ListArrayGenerator.from_form(form)

    elif isinstance(form, ak._v2.forms.IndexedForm):
        return IndexedArrayGenerator.from_form(form)

    elif isinstance(form, ak._v2.forms.IndexedOptionForm):
        return IndexedOptionArrayGenerator.from_form(form)

    elif isinstance(form, ak._v2.forms.ByteMaskedForm):
        return ByteMaskedArrayGenerator.from_form(form)

    elif isinstance(form, ak._v2.forms.BitMaskedForm):
        return BitMaskedArrayGenerator.from_form(form)

    elif isinstance(form, ak._v2.forms.UnmaskedForm):
        return UnmaskedArrayGenerator.from_form(form)

    elif isinstance(form, ak._v2.forms.RecordForm):
        return RecordArrayGenerator.from_form(form)

    elif isinstance(form, ak._v2.forms.UnionForm):
        return UnionArrayGenerator.from_form(form)

    else:
        raise AssertionError(f"unrecognized Form: {type(form)}")


class Generator:
    @classmethod
    def form_from_identifier(cls, form):
        if not form.has_identifier:
            return None
        else:
            raise NotImplementedError("TODO: identifiers in C++")

    def class_type_suffix(self, key):
        return (
            base64.encodebytes(struct.pack("q", hash(key)))
            .rstrip(b"=\n")
            .replace(b"+", b"")
            .replace(b"/", b"")
            .decode("ascii")
        )

    def _generate_common(self):
        params = [
            f"if (parameter == {json.dumps(name)}) return {json.dumps(json.dumps(value))};\n      "
            for name, value in self.parameters.items()
        ]

        return f"""
    const std::string parameter(const std::string& parameter) const noexcept {{
      {"" if len(params) == 0 else "".join(x for x in params)}return "null";
    }}

    value_type at(size_t at) const {{
      if (at >= stop_ - start_) {{
        throw std::out_of_range(std::to_string(at) + " is out of range");
      }}
      else {{
        return (*this)[at];
      }}
    }}
""".strip()

    def entry(self, length="length", ptrs="ptrs", flatlist_as_rvec=False):
        key = (self, flatlist_as_rvec)
        return f"awkward::{self.class_type(key[1:])}(0, {length}, 0, {ptrs})"


class NumpyArrayGenerator(Generator, ak._v2._lookup.NumpyLookup):
    @classmethod
    def from_form(cls, form):
        return NumpyArrayGenerator(
            form.primitive, cls.form_from_identifier(form), form.parameters
        )

    def __init__(self, primitive, identifier, parameters):
        self.primitive = primitive
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.primitive,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.primitive == other.primitive
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        return f"NumpyArray_{self.primitive}_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return {
            "bool": "bool",
            "int8": "int8_t",
            "uint8": "uint8_t",
            "int16": "int16_t",
            "uint16": "uint16_t",
            "int32": "int32_t",
            "uint32": "uint32_t",
            "int64": "int64_t",
            "uint64": "uint64_t",
            "float32": "float",
            "float64": "double",
            "complex64": "std::complex<float>",
            "complex128": "std::complex<double>",
            "datetime64": "std::time_point",
            "timedelta64": "std::duration",
        }[self.primitive]

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      return reinterpret_cast<{self.value_type(key[1:])}*>(ptrs_[which_ + {self.ARRAY}])[start_ + at];
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class RegularArrayGenerator(Generator, ak._v2._lookup.RegularLookup):
    @classmethod
    def from_form(cls, form):
        return RegularArrayGenerator(
            togenerator(form.content),
            form.size,
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, content, size, identifier, parameters):
        self.content = content
        self.size = size
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.content,
                self.size,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.content == other.content
            and self.size == other.size
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    @property
    def is_string(self):
        return isinstance(self.parameters, dict) and self.parameters.get(
            "__array__"
        ) in ("string", "bytestring")

    def is_flatlist(self):
        return isinstance(self.content, NumpyArrayGenerator)

    def class_type(self, key):
        return f"RegularArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return self.content.class_type(key)

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        if not self.is_string and not (flatlist_as_rvec and self.is_flatlist):
            self.content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            if self.is_string:
                out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef std::string value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t start = (start_ + at) * {self.size};
      ssize_t stop = start + {self.size};
      ssize_t which = ptrs_[which_ + {self.CONTENT}];
      char* content = reinterpret_cast<char*>(ptrs_[which + {NumpyArrayGenerator.ARRAY}]) + start;
      return value_type(content, stop - start);
    }}
  }};
}}
""".strip()
            elif flatlist_as_rvec and self.is_flatlist:
                nested_type = self.content.value_type(key[1:])
                value_type = f"ROOT::RVec<{nested_type}>"
                out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {value_type} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t start = (start_ + at) * {self.size};
      ssize_t stop = start + {self.size};
      ssize_t which = ptrs_[which_ + {self.CONTENT}];
      {nested_type}* content = reinterpret_cast<{nested_type}*>(ptrs_[which + {NumpyArrayGenerator.ARRAY}]) + start;
      return value_type(content, stop - start);
    }}
  }};
}}
""".strip()
            else:
                out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t start = (start_ + at) * {self.size};
      ssize_t stop = start + {self.size};
      return value_type(start, stop, ptrs_[which_ + {self.CONTENT}], ptrs_);
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class ListArrayGenerator(Generator, ak._v2._lookup.ListLookup):
    @classmethod
    def from_form(cls, form):
        if isinstance(form, ak._v2.forms.ListForm):
            index_string = form.starts
        else:
            index_string = form.offsets

        return ListArrayGenerator(
            index_string,
            togenerator(form.content),
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, index_type, content, identifier, parameters):
        if index_type == "i32":
            self.index_type = "int32_t"
        elif index_type == "u32":
            self.index_type = "uint32_t"
        elif index_type == "i64":
            self.index_type = "int64_t"
        else:
            raise AssertionError(index_type)
        self.content = content
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.index_type,
                self.content,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.index_type == other.index_type
            and self.content == other.content
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    @property
    def is_string(self):
        return isinstance(self.parameters, dict) and self.parameters.get(
            "__array__"
        ) in ("string", "bytestring")

    @property
    def is_flatlist(self):
        return isinstance(self.content, NumpyArrayGenerator)

    def class_type(self, key):
        return f"ListArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return self.content.class_type(key)

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        if not self.is_string and not (flatlist_as_rvec and self.is_flatlist):
            self.content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            if self.is_string:
                out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef std::string value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t start = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.STARTS}])[start_ + at];
      ssize_t stop = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.STOPS}])[start_ + at];
      ssize_t which = ptrs_[which_ + {self.CONTENT}];
      char* content = reinterpret_cast<char*>(ptrs_[which + {NumpyArrayGenerator.ARRAY}]) + start;
      return value_type(content, stop - start);
    }}
  }};
}}
""".strip()
            elif flatlist_as_rvec and self.is_flatlist:
                nested_type = self.content.value_type(key[1:])
                value_type = f"ROOT::RVec<{nested_type}>"
                out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {value_type} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t start = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.STARTS}])[start_ + at];
      ssize_t stop = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.STOPS}])[start_ + at];
      ssize_t which = ptrs_[which_ + {self.CONTENT}];
      {nested_type}* content = reinterpret_cast<{nested_type}*>(ptrs_[which + {NumpyArrayGenerator.ARRAY}]) + start;
      return value_type(content, stop - start);
    }}
  }};
}}
""".strip()
            else:
                out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t start = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.STARTS}])[start_ + at];
      ssize_t stop = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.STOPS}])[start_ + at];
      return value_type(start, stop, ptrs_[which_ + {self.CONTENT}], ptrs_);
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class IndexedArrayGenerator(Generator, ak._v2._lookup.IndexedLookup):
    @classmethod
    def from_form(cls, form):
        return IndexedArrayGenerator(
            form.index,
            togenerator(form.content),
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, index_type, content, identifier, parameters):
        if index_type == "i32":
            self.index_type = "int32_t"
        elif index_type == "u32":
            self.index_type = "uint32_t"
        elif index_type == "i64":
            self.index_type = "int64_t"
        else:
            raise AssertionError(index_type)
        self.content = content
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.index_type,
                self.content,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.index_type == other.index_type
            and self.content == other.content
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        return f"IndexedArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return self.content.value_type(key)

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        self.content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t index = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.INDEX}])[start_ + at];
      return {self.content.class_type(key[1:])}(index, index + 1, ptrs_[which_ + {self.CONTENT}], ptrs_)[0];
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class IndexedOptionArrayGenerator(Generator, ak._v2._lookup.IndexedOptionLookup):
    @classmethod
    def from_form(cls, form):
        return IndexedOptionArrayGenerator(
            form.index,
            togenerator(form.content),
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, index_type, content, identifier, parameters):
        if index_type == "i32":
            self.index_type = "int32_t"
        elif index_type == "i64":
            self.index_type = "int64_t"
        else:
            raise AssertionError(index_type)
        self.content = content
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.index_type,
                self.content,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.index_type == other.index_type
            and self.content == other.content
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        return f"IndexedOptionArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return f"std::optional<{self.content.value_type(key)}>"

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        self.content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      ssize_t index = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.INDEX}])[start_ + at];
      if (index >= 0) {{
        return value_type{{ {self.content.class_type(key[1:])}(index, index + 1, ptrs_[which_ + {self.CONTENT}], ptrs_)[0] }};
      }}
      else {{
        return std::nullopt;
      }}
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class ByteMaskedArrayGenerator(Generator, ak._v2._lookup.ByteMaskedLookup):
    @classmethod
    def from_form(cls, form):
        return ByteMaskedArrayGenerator(
            togenerator(form.content),
            form.valid_when,
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, content, valid_when, identifier, parameters):
        self.content = content
        self.valid_when = valid_when
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.content,
                self.valid_when,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.content == other.content
            and self.valid_when == other.valid_when
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        return f"ByteMaskedArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return f"std::optional<{self.content.value_type(key)}>"

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        self.content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      int8_t mask = reinterpret_cast<int8_t*>(ptrs_[which_ + {self.MASK}])[start_ + at];
      if ({"mask != 0" if self.valid_when else "mask == 0"}) {{
        return value_type{{ {self.content.class_type(key[1:])}(start_, stop_, ptrs_[which_ + {self.CONTENT}], ptrs_)[at] }};
      }}
      else {{
        return std::nullopt;
      }}
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class BitMaskedArrayGenerator(Generator, ak._v2._lookup.BitMaskedLookup):
    @classmethod
    def from_form(cls, form):
        return BitMaskedArrayGenerator(
            togenerator(form.content),
            form.valid_when,
            form.lsb_order,
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, content, valid_when, lsb_order, identifier, parameters):
        self.content = content
        self.valid_when = valid_when
        self.lsb_order = lsb_order
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                self.content,
                self.valid_when,
                self.lsb_order,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.content == other.content
            and self.valid_when == other.valid_when
            and self.lsb_order == other.lsb_order
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        return f"BitMaskedArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return f"std::optional<{self.content.value_type(key)}>"

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        self.content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      size_t startat = start_ + at;
      size_t bitat = startat / 8;
      size_t shift = startat % 8;
      uint8_t byte = reinterpret_cast<uint8_t*>(ptrs_[which_ + {self.MASK}])[bitat];
      uint8_t mask = {"(byte >> shift) & 1" if self.lsb_order else "(byte << shift) & 128"};

      if ({"mask != 0" if self.valid_when else "mask == 0"}) {{
        return value_type{{ {self.content.class_type(key[1:])}(start_, stop_, ptrs_[which_ + {self.CONTENT}], ptrs_)[at] }};
      }}
      else {{
        return std::nullopt;
      }}
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class UnmaskedArrayGenerator(Generator, ak._v2._lookup.UnmaskedLookup):
    @classmethod
    def from_form(cls, form):
        return UnmaskedArrayGenerator(
            togenerator(form.content), cls.form_from_identifier(form), form.parameters
        )

    def __init__(self, content, identifier, parameters):
        self.content = content
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (type(self), self.content, self.identifier, json.dumps(self.parameters))
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.content == other.content
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        return f"UnmaskedArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return f"std::optional<{self.content.value_type(key)}>"

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        self.content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      return value_type{{ {self.content.class_type(key[1:])}(start_, stop_, ptrs_[which_ + {self.CONTENT}], ptrs_)[at] }};
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class RecordGenerator(Generator, ak._v2._lookup.RecordLookup):
    def __init__(self, contents, fields, parameters):
        self.contents = tuple(contents)
        self.fields = None if fields is None else tuple(fields)
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.contents,
                self.fields,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.contents == other.contents
            and self.fields == other.fields
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        if (
            isinstance(self.parameters, dict)
            and self.parameters.get("__record__") is not None
        ):
            insert = "_" + self.parameters["__record__"]
        else:
            insert = ""
        return f"Record{insert}_{self.class_type_suffix((self,) + key)}"

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_RecordView(compiler, use_cached=use_cached)
        for content in self.contents:
            content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            params = [
                f"if (parameter == {json.dumps(name)}) return {json.dumps(json.dumps(value))};\n      "
                for name, value in self.parameters.items()
            ]

            if self.fields is None:
                fieldnames = [f"slot{i}" for i in range(len(self.contents))]
            else:
                fieldnames = self.fields
            getfields = []
            for i, (fieldname, content) in enumerate(zip(fieldnames, self.contents)):
                if re.match("^[A-Za-z_][A-Za-z_0-9]*$", fieldname) is not None:
                    getfields.append(
                        f"""
    {content.value_type(key[1:])} {fieldname}() const noexcept {{
      return {content.class_type(key[1:])}(at_, at_ + 1, ptrs_[which_ + {self.CONTENTS + i}], ptrs_)[0];
    }}
""".strip()
                    )

            eoln = "\n    "
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public RecordView {{
  public:
    {self.class_type(key[1:])}(ssize_t at, ssize_t which, ssize_t* ptrs)
      : RecordView(at, which, ptrs) {{ }}

    const std::string parameter(const std::string& parameter) const noexcept {{
      {"" if len(params) == 0 else "".join(x for x in params)}return "null";
    }}

    {eoln.join(getfields)}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class RecordArrayGenerator(Generator, ak._v2._lookup.RecordLookup):
    @classmethod
    def from_form(cls, form):
        return RecordArrayGenerator(
            [togenerator(x) for x in form.contents],
            None if form.is_tuple else form.fields,
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, contents, fields, identifier, parameters):
        self.contents = tuple(contents)
        self.fields = None if fields is None else tuple(fields)
        self.identifier = identifier
        self.parameters = parameters

        self.record = RecordGenerator(contents, fields, parameters)

    def __hash__(self):
        return hash(
            (
                type(self),
                self.contents,
                self.fields,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.contents == other.contents
            and self.fields == other.fields
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        if (
            isinstance(self.parameters, dict)
            and self.parameters.get("__record__") is not None
        ):
            insert = "_" + self.parameters["__record__"]
        else:
            insert = ""
        return f"RecordArray{insert}_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return self.record.class_type(key)

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        self.record.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      return value_type(start_ + at, which_, ptrs_);
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)


class UnionArrayGenerator(Generator, ak._v2._lookup.UnionLookup):
    @classmethod
    def from_form(cls, form):
        return UnionArrayGenerator(
            form.index,
            [togenerator(x) for x in form.contents],
            cls.form_from_identifier(form),
            form.parameters,
        )

    def __init__(self, index_type, contents, identifier, parameters):
        if index_type == "i32":
            self.index_type = "int32_t"
        elif index_type == "u32":
            self.index_type = "uint32_t"
        elif index_type == "i64":
            self.index_type = "int64_t"
        else:
            raise AssertionError(index_type)
        self.contents = tuple(contents)
        self.identifier = identifier
        self.parameters = parameters

    def __hash__(self):
        return hash(
            (
                type(self),
                self.index_type,
                self.contents,
                self.identifier,
                json.dumps(self.parameters),
            )
        )

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.index_type == other.index_type
            and self.contents == other.contents
            and self.identifier == other.identifier
            and self.parameters == other.parameters
        )

    def class_type(self, key):
        return f"UnionArray_{self.class_type_suffix((self,) + key)}"

    def value_type(self, key):
        return f"std::variant<{','.join(x.value_type(key) for x in self.contents)}>"

    def generate(self, compiler, use_cached=True, flatlist_as_rvec=False):
        generate_ArrayView(compiler, use_cached=use_cached)
        for content in self.contents:
            content.generate(compiler, use_cached, flatlist_as_rvec)

        key = (self, flatlist_as_rvec)
        if not use_cached or key not in cache:
            cases = []
            for i, content in enumerate(self.contents):
                cases.append(
                    f"""
        case {i}:
          return value_type{{ {content.class_type(key[1:])}(index, index + 1, ptrs_[which_ + {self.CONTENTS + i}], ptrs_)[0] }};
""".strip()
                )

            eoln = "\n        "
            out = f"""
namespace awkward {{
  class {self.class_type(key[1:])}: public ArrayView {{
  public:
    {self.class_type(key[1:])}(ssize_t start, ssize_t stop, ssize_t which, ssize_t* ptrs)
      : ArrayView(start, stop, which, ptrs) {{ }}

    typedef {self.value_type(key[1:])} value_type;

    {self._generate_common()}

    value_type operator[](size_t at) const noexcept {{
      int8_t tag = reinterpret_cast<int8_t*>(ptrs_[which_ + {self.TAGS}])[start_ + at];
      {self.index_type} index = reinterpret_cast<{self.index_type}*>(ptrs_[which_ + {self.INDEX}])[start_ + at];
      switch (tag) {{
        {eoln.join(cases)}
        default:
          return value_type{{ {self.contents[0].class_type(key[1:])}(index, index + 1, ptrs_[which_ + {self.CONTENTS + 0}], ptrs_)[0] }};
      }}
    }}
  }};
}}
""".strip()
            cache[key] = out
            compiler(out)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def to_arrow(
    array,
    list_to32=False,
    string_to32=False,
    bytestring_to32=False,
    emptyarray_to=None,
    categorical_as_dictionary=False,
    extensionarray=True,
    count_nulls=True,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        list_to32 (bool): If True, convert Awkward lists into 32-bit Arrow lists
            if they're small enough, even if it means an extra conversion. Otherwise,
            signed 32-bit #ak.types.ListType maps to Arrow `ListType`,
            signed 64-bit #ak.types.ListType maps to Arrow `LargeListType`,
            and unsigned 32-bit #ak.types.ListType picks whichever Arrow type its
            values fit into.
        string_to32 (bool): Same as the above for Arrow `string` and `large_string`.
        bytestring_to32 (bool): Same as the above for Arrow `binary` and `large_binary`.
        emptyarray_to (None or dtype): If None, #ak.types.UnknownType maps to Arrow's
            null type; otherwise, it is converted a given numeric dtype.
        categorical_as_dictionary (bool): If True, #ak.layout.IndexedArray and
            #ak.layout.IndexedOptionArray labeled with `__array__ = "categorical"`
            are mapped to Arrow `DictionaryArray`; otherwise, the projection is
            evaluated before conversion (always the case without
            `__array__ = "categorical"`).
        extensionarray (bool): If True, this function returns extended Arrow arrays
            (at all levels of nesting), which preserve metadata so that Awkward \u2192
            Arrow \u2192 Awkward preserves the array's #ak.types.Type (though not
            the #ak.forms.Form). If False, this function returns generic Arrow arrays
            that might be needed for third-party tools that don't recognize Arrow's
            extensions. Even with `extensionarray=False`, the values produced by
            Arrow's `to_pylist` method are the same as the values produced by Awkward's
            #ak.to_list.
        count_nulls (bool): If True, count the number of missing values at each level
            and include these in the resulting Arrow array, which makes some downstream
            applications faster. If False, skip the up-front cost of counting them.

    Converts an Awkward Array into an Apache Arrow array.

    This produces arrays of type `pyarrow.Array`. You might need to further
    manipulations (using the pyarrow library) to build a `pyarrow.ChunkedArray`,
    a `pyarrow.RecordBatch`, or a `pyarrow.Table`. For the latter, see #ak.to_arrow_table.

    This function always preserves the values of a dataset; i.e. the Python objects
    returned by #ak.to_list are identical to the Python objects returned by Arrow's
    `to_pylist` method. With `extensionarray=True`, this function also preserves the
    data type (high-level #ak.types.Type, though not the low-level #ak.forms.Form),
    even through Parquet, making Parquet a good way to save Awkward Arrays for later
    use. If any third-party tools don't recognize Arrow's extension arrays, set this
    option to False for plain Arrow arrays.

    See also #ak.from_arrow, #ak.to_arrow_table, #ak.to_parquet, #ak.from_arrow_schema.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=False, allow_other=False
    )

    return layout.to_arrow(
        list_to32=list_to32,
        string_to32=string_to32,
        bytestring_to32=bytestring_to32,
        emptyarray_to=emptyarray_to,
        categorical_as_dictionary=categorical_as_dictionary,
        extensionarray=extensionarray,
        count_nulls=count_nulls,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def unzip(array, highlevel=True, behavior=None):
    raise NotImplementedError

    # """
    # Args:
    #     array: Array to unzip into individual fields.
    #     highlevel (bool): If True, return an #ak.Array; otherwise, return
    #         a low-level #ak.layout.Content subclass.
    #     behavior (None or dict): Custom #ak.behavior for the output array, if
    #         high-level.

    # If the `array` contains tuples or records, this operation splits them
    # into a Python tuple of arrays, one for each field.

    # If the `array` does not contain tuples or records, the single `array`
    # is placed in a length 1 Python tuple.

    # For example,

    #     >>> array = ak.Array([{"x": 1.1, "y": [1]},
    #     ...                   {"x": 2.2, "y": [2, 2]},
    #     ...                   {"x": 3.3, "y": [3, 3, 3]}])
    #     >>> x, y = ak.unzip(array)
    #     >>> x
    #     <Array [1.1, 2.2, 3.3] type='3 * float64'>
    #     >>> y
    #     <Array [[1], [2, 2], [3, 3, 3]] type='3 * var * int64'>
    # """
    # behavior = ak._util.behaviorof(array, behavior=behavior)
    # layout = ak.operations.convert.to_layout(array, allow_record=True, allow_other=False)
    # fields = ak.operations.describe.fields(layout)

    ### FIXME: In v2, you can use return_array=False in Content.recursively_apply
    ###        to perform an action on the whole array but not return a result.
    ###        No need for (brittle) specialized code like the following.

    # def check_for_union(layout):
    #     if isinstance(layout, ak.partition.PartitionedArray):
    #         for x in layout.partitions:
    #             check_for_union(x)

    #     elif isinstance(layout, ak.layout.RecordArray):
    #         pass  # don't descend into nested records

    #     elif isinstance(layout, ak.layout.Record):
    #         pass  # don't descend into nested records

    #     elif isinstance(
    #         layout,
    #         (
    #             ak.layout.UnionArray8_32,
    #             ak.layout.UnionArray8_U32,
    #             ak.layout.UnionArray8_64,
    #         ),
    #     ):
    #         for content in layout.contents:
    #             if set(ak.operations.describe.fields(content)) != set(fields):
    #                 raise ValueError("union of different sets of fields, cannot ak.unzip")

    #     elif hasattr(layout, "content"):
    #         check_for_union(layout.content)

    # check_for_union(layout)

    # if len(fields) == 0:
    #     return (ak._util.maybe_wrap(layout, behavior, highlevel),)
    # else:
    #     return tuple(ak._util.maybe_wrap(layout[n], behavior, highlevel) for n in fields)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

to_list = ak._v2.operations.convert.to_list


def test():
    def _apply_ufunc(ufunc, method, inputs, kwargs):
        nextinputs = []
        for x in inputs:
            if (
                isinstance(x, ak._v2.highlevel.Array)
                and x.layout.is_IndexedType
                and not x.layout.is_OptionType
            ):
                nextinputs.append(
                    ak._v2.highlevel.Array(
                        x.layout.project(), behavior=ak._v2._util.behavior_of(x)
                    )
                )
            else:
                nextinputs.append(x)

        return getattr(ufunc, method)(*nextinputs, **kwargs)

    ak._v2.behavior[np.ufunc, "categorical"] = _apply_ufunc

    array = ak._v2.highlevel.Array(
        ak._v2.contents.IndexedArray(
            ak._v2.index.Index64(np.array([0, 1, 2, 1, 3, 1, 4])),
            ak._v2.contents.NumpyArray(np.array([321, 1.1, 123, 999, 2])),
            parameters={"__array__": "categorical"},
        )
    )
    assert to_list(array * 10) == [3210, 11, 1230, 11, 9990, 11, 20]

    array = ak._v2.highlevel.Array(["HAL"])
    with pytest.raises(TypeError):
        array + 1



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


def with_parameter(array, parameter, value, highlevel=True, behavior=None):

    """
    Args:
        array: Data convertible into an Awkward Array.
        parameter (str): Name of the parameter to set on that array.
        value (JSON): Value of the parameter to set on that array.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.layout.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    This function returns a new array with a parameter set on the outermost
    node of its #ak.Array.layout.

    Note that a "new array" is a lightweight shallow copy, not a duplication
    of large data buffers.

    You can also remove a single parameter with this function, since setting
    a parameter to None is equivalent to removing it.
    """
    layout = ak._v2.operations.convert.to_layout(
        array, allow_record=True, allow_other=False
    )

    out = layout.withparameter(parameter, value)

    return ak._v2._util.wrap(
        out, ak._v2._util.behavior_of(array, behavior=behavior), highlevel
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


# @ak._v2._connect.numpy.implements("broadcast_arrays")
def broadcast_arrays(*arrays, **kwargs):
    raise NotImplementedError


#     """
#     Args:
#         arrays: Arrays to broadcast into the same structure.
#         left_broadcast (bool): If True, follow rules for implicit
#             left-broadcasting, as described below.
#         right_broadcast (bool): If True, follow rules for implicit
#             right-broadcasting, as described below.
#         highlevel (bool, default is True): If True, return an #ak.Array;
#             otherwise, return a low-level #ak.layout.Content subclass.

#     Like NumPy's
#     [broadcast_arrays](https://docs.scipy.org/doc/numpy/reference/generated/numpy.broadcast_arrays.html)
#     function, this function returns the input `arrays` with enough elements
#     duplicated that they can be combined element-by-element.

#     For NumPy arrays, this means that scalars are replaced with arrays with
#     the same scalar value repeated at every element of the array, and regular
#     dimensions are created to increase low-dimensional data into
#     high-dimensional data.

#     For example,

#         >>> ak.broadcast_arrays(5,
#         ...                     [1, 2, 3, 4, 5])
#         [<Array [5, 5, 5, 5, 5] type='5 * int64'>,
#          <Array [1, 2, 3, 4, 5] type='5 * int64'>]

#     and

#         >>> ak.broadcast_arrays(np.array([1, 2, 3]),
#         ...                     np.array([[0.1, 0.2, 0.3], [10, 20, 30]]))
#         [<Array [[  1,   2,   3], [ 1,  2,  3]] type='2 * 3 * int64'>,
#          <Array [[0.1, 0.2, 0.3], [10, 20, 30]] type='2 * 3 * float64'>]

#     Note that in the second example, when the `3 * int64` array is expanded
#     to match the `2 * 3 * float64` array, it is the deepest dimension that
#     is aligned. If we try to match a `2 * int64` with the `2 * 3 * float64`,

#         >>> ak.broadcast_arrays(np.array([1, 2]),
#         ...                     np.array([[0.1, 0.2, 0.3], [10, 20, 30]]))
#         ValueError: cannot broadcast RegularArray of size 2 with RegularArray of size 3

#     NumPy has the same behavior: arrays with different numbers of dimensions
#     are aligned to the right before expansion. One can control this by
#     explicitly adding a new axis (reshape to add a dimension of length 1)
#     where the expansion is supposed to take place because a dimension of
#     length 1 can be expanded like a scalar.

#         >>> ak.broadcast_arrays(np.array([1, 2])[:, np.newaxis],
#         ...                     np.array([[0.1, 0.2, 0.3], [10, 20, 30]]))
#         [<Array [[  1,   1,   1], [ 2,  2,  2]] type='2 * 3 * int64'>,
#          <Array [[0.1, 0.2, 0.3], [10, 20, 30]] type='2 * 3 * float64'>]

#     Again, NumPy does the same thing (`np.newaxis` is equal to None, so this
#     trick is often shown with None in the slice-tuple). Where the broadcasting
#     happens can be controlled, but numbers of dimensions that don't match are
#     implicitly aligned to the right (fitting innermost structure, not
#     outermost).

#     While that might be an arbitrary decision for rectilinear arrays, it is
#     much more natural for implicit broadcasting to align left for tree-like
#     structures. That is, the root of each data structure should agree and
#     leaves may be duplicated to match. For example,

#         >>> ak.broadcast_arrays([            100,   200,        300],
#         ...                     [[1.1, 2.2, 3.3],    [], [4.4, 5.5]])
#         [<Array [[100, 100, 100], [], [300, 300]] type='3 * var * int64'>,
#          <Array [[1.1, 2.2, 3.3], [], [4.4, 5.5]] type='3 * var * float64'>]

#     One typically wants single-item-per-element data to be duplicated to
#     match multiple-items-per-element data. Operations on the broadcasted
#     arrays like

#         one_dimensional + nested_lists

#     would then have the same effect as the procedural code

#         for x, outer in zip(one_dimensional, nested_lists):
#             output = []
#             for inner in outer:
#                 output.append(x + inner)
#             yield output

#     where `x` has the same value for each `inner` in the inner loop.

#     Awkward Array's broadcasting manages to have it both ways by applying the
#     following rules:

#        * If all dimensions are regular (i.e. #ak.types.RegularType), like NumPy,
#          implicit broadcasting aligns to the right, like NumPy.
#        * If any dimension is variable (i.e. #ak.types.ListType), which can
#          never be true of NumPy, implicit broadcasting aligns to the left.
#        * Explicit broadcasting with a length-1 regular dimension always
#          broadcasts, like NumPy.

#     Thus, it is important to be aware of the distinction between a dimension
#     that is declared to be regular in the type specification and a dimension
#     that is allowed to be variable (even if it happens to have the same length
#     for all elements). This distinction is can be accessed through the
#     #ak.Array.type, but it is lost when converting an array into JSON or
#     Python objects.
#     """
#     (highlevel, left_broadcast, right_broadcast) = ak._v2._util.extra(
#         (),
#         kwargs,
#         [("highlevel", True), ("left_broadcast", True), ("right_broadcast", True)],
#     )

#     inputs = []
#     for x in arrays:
#         y = ak._v2.operations.convert.to_layout(x, allow_record=True, allow_other=True)
#         if isinstance(y, ak.partition.PartitionedArray):   # NO PARTITIONED ARRAY
#             y = y.toContent()
#         if not isinstance(y, (ak._v2.contents.Content, ak._v2.contents.Record)):
#             y = ak._v2.contents.NumpyArray(ak.nplike.of(*arrays).array([y]))
#         inputs.append(y)

#     def getfunction(inputs):
#         if all(isinstance(x, ak._v2.contents.NumpyArray) for x in inputs):
#             return lambda: tuple(inputs)
#         else:
#             return None

#     behavior = ak._v2._util.behaviorof(*arrays)
#     out = ak._v2._util.broadcast_and_apply(
#         inputs,
#         getfunction,
#         behavior,
#         left_broadcast=left_broadcast,
#         right_broadcast=right_broadcast,
#         pass_depth=False,
#         numpy_to_regular=True,
#     )
#     assert isinstance(out, tuple)
#     if highlevel:
#         return [ak._v2._util.wrap(x, behavior) for x in out]
#     else:
#         return list(out)
